@@ -19,18 +19,122 @@ static void *mcp_profile_alloc(void *ud, void *ptr, size_t osize, size_t nsize);
 /******** EXTERNAL FUNCTIONS ******/
 // functions starting with _ are breakouts for the public functions.
 
-// We want to ensure we slowly iterate through VM memory in two cases:
-// - If using the newer zero-allocation API, old backends will never close
-// after reload.
-// - If the server is completely idle and user is issuing reloads, they will
-// be confused when old backends do not close.
-// So every few seconds we push one minimal GC step.
-static void proxy_gc_poke(evutil_socket_t fd, short event, void *arg) {
-    LIBEVENT_THREAD *t = arg;
+static inline void _proxy_advance_lastkb(lua_State *L, LIBEVENT_THREAD *t) {
+    int new_kb = lua_gc(L, LUA_GCCOUNT);
+    // We need to slew the increase in "gc pause" because the lua GC actually
+    // needs to run twice to free a userdata: once to run the _gc's and again
+    // to actually clean up the object.
+    // Meaning we will continually increase in size.
+    if (new_kb > t->proxy_vm_last_kb) {
+        new_kb = t->proxy_vm_last_kb + (new_kb - t->proxy_vm_last_kb) * 0.50;
+    }
+
+    // remove the memory freed during this cycle so we can kick off the GC
+    // early if we're very aggressively making garbage.
+    // carry our negative delta forward so a huge reclaim can push for a
+    // couple cycles.
+    if (t->proxy_vm_negative_delta >= new_kb) {
+        t->proxy_vm_negative_delta -= new_kb;
+        new_kb = 1;
+    } else {
+        new_kb -= t->proxy_vm_negative_delta;
+        t->proxy_vm_negative_delta = 0;
+    }
+
+    t->proxy_vm_last_kb = new_kb;
+}
+
+// The lua GC is paused while running requests. Run it manually inbetween
+// processing network events.
+void proxy_gc_poke(LIBEVENT_THREAD *t) {
     lua_State *L = t->L;
+    struct proxy_int_stats *is = t->proxy_int_stats;
+    int vm_kb = lua_gc(L, LUA_GCCOUNT) + t->proxy_vm_extra_kb;
+    if (t->proxy_vm_last_kb == 0) {
+        t->proxy_vm_last_kb = vm_kb;
+    }
+    WSTAT_L(t);
+    is->vm_memory_kb = vm_kb;
+    WSTAT_UL(t);
+
+    // equivalent of luagc "pause" value
+    int last = t->proxy_vm_last_kb;
+    if (t->proxy_vm_gcrunning <= 0 && vm_kb > last * 2) {
+        t->proxy_vm_gcrunning = 1;
+        //fprintf(stderr, "PROXYGC: proxy_gc_poke START [cur: %d - last: %d]\n", vm_kb, last);
+    }
+
+    // We configure small GC "steps" then increase the number of times we run
+    // a step based on current memory usage.
+    if (t->proxy_vm_gcrunning > 0) {
+        t->proxy_vm_needspoke = false;
+        int loops = t->proxy_vm_gcrunning;
+        int done = 0;
+        /*fprintf(stderr, "PROXYGC: proxy_gc_poke [cur: %d - last: %d - loops: %d]\n",
+            vm_kb,
+            t->proxy_vm_last_kb,
+            loops);*/
+        while (loops-- && !done) {
+            // reset counters once full GC cycle has completed
+            done = lua_gc(L, LUA_GCSTEP, 0);
+        }
+
+        int vm_kb_after = lua_gc(L, LUA_GCCOUNT);
+        int vm_kb_clean = vm_kb - t->proxy_vm_extra_kb;
+        if (vm_kb_clean > vm_kb_after) {
+            // track the amount of memory freed during the GC cycle.
+            t->proxy_vm_negative_delta += vm_kb_clean - vm_kb_after;
+        }
+
+        if (done) {
+            _proxy_advance_lastkb(L, t);
+            t->proxy_vm_extra_kb = 0;
+            t->proxy_vm_gcrunning = 0;
+            WSTAT_L(t);
+            is->vm_gc_runs++;
+            WSTAT_UL(t);
+            //fprintf(stderr, "PROXYGC: proxy_gc_poke COMPLETE [cur: %d next: %d]\n", lua_gc(L, LUA_GCCOUNT), t->proxy_vm_last_kb);
+        }
+
+        // increase the aggressiveness by memory bloat level.
+        if (t->proxy_vm_gcrunning && (last*2) + (last * t->proxy_vm_gcrunning*0.25) < vm_kb) {
+            t->proxy_vm_gcrunning++;
+            //fprintf(stderr, "PROXYGC: proxy_gc_poke INCREASING AGGRESSIVENESS [cur: %d - aggro: %d]\n", t->proxy_vm_last_kb, t->proxy_vm_gcrunning);
+        } else if (t->proxy_vm_gcrunning > 1) {
+            // memory can drop during a run, let the GC slow down again.
+            t->proxy_vm_gcrunning--;
+            //fprintf(stderr, "PROXYGC: proxy_gc_poke DECREASING AGGRESSIVENESS [cur: %d - aggro: %d]\n", t->proxy_vm_last_kb, t->proxy_vm_gcrunning);
+        }
+    }
+}
+
+// every couple seconds we force-run one GC step.
+// this is needed until after API1 is retired and pool objects are no longer
+// managed by the GC.
+// We use a negative value so a "timer poke" GC run doesn't cause requests to
+// suddenly aggressively run the GC.
+static void proxy_gc_timerpoke(evutil_socket_t fd, short event, void *arg) {
+    LIBEVENT_THREAD *t = arg;
     struct timeval next = { PROXY_GC_BACKGROUND_SECONDS, 0 };
-    lua_gc(L, LUA_GCSTEP, 0);
     evtimer_add(t->proxy_gc_timer, &next);
+    // if GC ran within the last few seconds, don't do anything.
+    if (!t->proxy_vm_needspoke) {
+        t->proxy_vm_needspoke = true;
+        return;
+    }
+
+    // if we weren't told to skip and there's otherwise no GC running, start a
+    // GC run.
+    if (t->proxy_vm_gcrunning == 0) {
+        t->proxy_vm_gcrunning = -1;
+    }
+
+    // only advance GC if we're doing our own timer run.
+    if (t->proxy_vm_gcrunning == -1 && lua_gc(t->L, LUA_GCSTEP, 0)) {
+        _proxy_advance_lastkb(t->L, t);
+        t->proxy_vm_extra_kb = 0;
+        t->proxy_vm_gcrunning = 0;
+    }
 }
 
 bool proxy_bufmem_checkadd(LIBEVENT_THREAD *t, int len) {
@@ -46,22 +150,25 @@ bool proxy_bufmem_checkadd(LIBEVENT_THREAD *t, int len) {
 }
 
 // see also: process_extstore_stats()
-void proxy_stats(void *arg, ADD_STAT add_stats, conn *c) {
+void proxy_stats(void *arg, ADD_STAT add_stats, void *c) {
     if (arg == NULL) {
        return;
     }
     proxy_ctx_t *ctx = arg;
-    STAT_L(ctx);
 
+    STAT_L(ctx);
     APPEND_STAT("proxy_config_reloads", "%llu", (unsigned long long)ctx->global_stats.config_reloads);
     APPEND_STAT("proxy_config_reload_fails", "%llu", (unsigned long long)ctx->global_stats.config_reload_fails);
+    APPEND_STAT("proxy_config_cron_runs", "%llu", (unsigned long long)ctx->global_stats.config_cron_runs);
+    APPEND_STAT("proxy_config_cron_fails", "%llu", (unsigned long long)ctx->global_stats.config_cron_fails);
     APPEND_STAT("proxy_backend_total", "%llu", (unsigned long long)ctx->global_stats.backend_total);
     APPEND_STAT("proxy_backend_marked_bad", "%llu", (unsigned long long)ctx->global_stats.backend_marked_bad);
     APPEND_STAT("proxy_backend_failed", "%llu", (unsigned long long)ctx->global_stats.backend_failed);
+    APPEND_STAT("proxy_request_failed_depth", "%llu", (unsigned long long)ctx->global_stats.request_failed_depth);
     STAT_UL(ctx);
 }
 
-void process_proxy_stats(void *arg, ADD_STAT add_stats, conn *c) {
+void process_proxy_stats(void *arg, ADD_STAT add_stats, void *c) {
     char key_str[STAT_KEY_LEN];
     struct proxy_int_stats istats = {0};
     uint64_t req_limit = 0;
@@ -77,8 +184,9 @@ void process_proxy_stats(void *arg, ADD_STAT add_stats, conn *c) {
     buffer_memory_limit = ctx->buffer_memory_limit;
 
     // prepare aggregated counters.
-    struct proxy_user_stats *us = &ctx->user_stats;
-    uint64_t counters[us->num_stats];
+    struct proxy_user_stats_entry *us = ctx->user_stats;
+    int stats_num = ctx->user_stats_num;
+    uint64_t counters[stats_num];
     memset(counters, 0, sizeof(counters));
 
     // TODO (v3): more globals to remove and/or change API method.
@@ -91,8 +199,10 @@ void process_proxy_stats(void *arg, ADD_STAT add_stats, conn *c) {
         for (int i = 0; i < CMD_FINAL; i++) {
             istats.counters[i] += is->counters[i];
         }
-        if (tus && tus->num_stats >= us->num_stats) {
-            for (int i = 0; i < us->num_stats; i++) {
+        istats.vm_gc_runs += is->vm_gc_runs;
+        istats.vm_memory_kb += is->vm_memory_kb;
+        if (tus && tus->num_stats >= stats_num) {
+            for (int i = 0; i < stats_num; i++) {
                 counters[i] += tus->counters[i];
             }
         }
@@ -103,10 +213,31 @@ void process_proxy_stats(void *arg, ADD_STAT add_stats, conn *c) {
     }
 
     // return all of the user generated stats
-    for (int x = 0; x < us->num_stats; x++) {
-        if (us->names[x] && us->names[x][0]) {
-            snprintf(key_str, STAT_KEY_LEN-1, "user_%s", us->names[x]);
-            APPEND_STAT(key_str, "%llu", (unsigned long long)counters[x]);
+    if (ctx->user_stats_namebuf) {
+        char vbuf[INCR_MAX_STORAGE_LEN];
+        char *e = NULL; // ptr into vbuf
+        const char *pfx = "user_";
+        const size_t pfxlen = strlen(pfx);
+        for (int x = 0; x < stats_num; x++) {
+            if (us[x].cname) {
+                char *name = ctx->user_stats_namebuf + us[x].cname;
+                size_t nlen = strlen(name);
+                if (nlen > STAT_KEY_LEN-6) {
+                    // impossible, but for paranoia.
+                    nlen = STAT_KEY_LEN-6;
+                }
+                // avoiding an snprintf call for some performance ("user_%s")
+                memcpy(key_str, pfx, pfxlen);
+                memcpy(key_str+pfxlen, name, nlen);
+                key_str[pfxlen+nlen] = '\0';
+
+                // APPEND_STAT() calls another snprintf, which calls our
+                // add_stats argument. Lets skip yet another snprintf with
+                // some unrolling.
+                e = itoa_u64(counters[x], vbuf);
+                *(e+1) = '\0';
+                add_stats(key_str, pfxlen+nlen, vbuf, e-vbuf, c);
+            }
         }
     }
 
@@ -127,6 +258,8 @@ void process_proxy_stats(void *arg, ADD_STAT add_stats, conn *c) {
     APPEND_STAT("active_req_limit", "%llu", (unsigned long long)req_limit);
     APPEND_STAT("buffer_memory_limit", "%llu", (unsigned long long)buffer_memory_limit);
     APPEND_STAT("buffer_memory_used", "%llu", (unsigned long long)buffer_memory_used);
+    APPEND_STAT("vm_gc_runs", "%llu", (unsigned long long)istats.vm_gc_runs);
+    APPEND_STAT("vm_memory_kb", "%llu", (unsigned long long)istats.vm_memory_kb);
     APPEND_STAT("cmd_mg", "%llu", (unsigned long long)istats.counters[CMD_MG]);
     APPEND_STAT("cmd_ms", "%llu", (unsigned long long)istats.counters[CMD_MS]);
     APPEND_STAT("cmd_md", "%llu", (unsigned long long)istats.counters[CMD_MD]);
@@ -149,7 +282,7 @@ void process_proxy_stats(void *arg, ADD_STAT add_stats, conn *c) {
     APPEND_STAT("cmd_replace", "%llu", (unsigned long long)istats.counters[CMD_REPLACE]);
 }
 
-void process_proxy_funcstats(void *arg, ADD_STAT add_stats, conn *c) {
+void process_proxy_funcstats(void *arg, ADD_STAT add_stats, void *c) {
     char key_str[STAT_KEY_LEN];
     if (!arg) {
         return;
@@ -177,8 +310,39 @@ void process_proxy_funcstats(void *arg, ADD_STAT add_stats, conn *c) {
             snprintf(key_str, STAT_KEY_LEN-1, "slots_%s", name);
             APPEND_STAT(key_str, "%d", slots);
         } else {
-            // TODO: Is it safe to delete keys in the middle here?
-            // not worried at all about just leaking memory here.
+            // TODO: It is safe to delete keys here. Slightly complex so low
+            // priority.
+        }
+    }
+
+    pthread_mutex_unlock(&ctx->sharedvm_lock);
+}
+
+void process_proxy_bestats(void *arg, ADD_STAT add_stats, void *c) {
+    char key_str[STAT_KEY_LEN];
+    if (!arg) {
+        return;
+    }
+    proxy_ctx_t *ctx = arg;
+    lua_State *L = ctx->proxy_sharedvm;
+    pthread_mutex_lock(&ctx->sharedvm_lock);
+
+    // iterate all of the listed backends
+    lua_pushnil(L);
+    while (lua_next(L, SHAREDVM_BACKEND_IDX) != 0) {
+        int n = lua_tointeger(L, -1);
+        lua_pop(L, 1); // drop the value, leave the key.
+        if (n != 0) {
+            // now grab the name key.
+            const char *name = lua_tostring(L, -1);
+            snprintf(key_str, STAT_KEY_LEN-1, "bad_%s", name);
+            APPEND_STAT(key_str, "%d", n);
+        } else {
+            // delete keys of backends that are no longer bad or no longer
+            // exist to keep the table small.
+            const char *name = lua_tostring(L, -1);
+            lua_pushnil(L);
+            lua_setfield(L, SHAREDVM_BACKEND_IDX, name);
         }
     }
 
@@ -189,7 +353,7 @@ void process_proxy_funcstats(void *arg, ADD_STAT add_stats, conn *c) {
 void *proxy_init(bool use_uring, bool proxy_memprofile) {
     proxy_ctx_t *ctx = calloc(1, sizeof(proxy_ctx_t));
     ctx->use_uring = use_uring;
-    ctx->memprofile = proxy_memprofile; // FIXME: take from argument.
+    ctx->memprofile = proxy_memprofile;
 
     pthread_mutex_init(&ctx->config_lock, NULL);
     pthread_cond_init(&ctx->config_cond, NULL);
@@ -210,8 +374,10 @@ void *proxy_init(bool use_uring, bool proxy_memprofile) {
     ctx->tunables.read.tv_sec = 3;
     ctx->tunables.flap_backoff_ramp = 1.5;
     ctx->tunables.flap_backoff_max = 3600;
+    ctx->tunables.backend_depth_limit = 0;
     ctx->tunables.max_ustats = MAX_USTATS_DEFAULT;
     ctx->tunables.use_iothread = false;
+    ctx->tunables.use_tls = false;
 
     STAILQ_INIT(&ctx->manager_head);
     lua_State *L = NULL;
@@ -226,6 +392,10 @@ void *proxy_init(bool use_uring, bool proxy_memprofile) {
     luaL_openlibs(L);
     // NOTE: might need to differentiate the libs yes?
     proxy_register_libs(ctx, NULL, L);
+    // Create the cron table.
+    lua_newtable(L);
+    ctx->cron_ref = luaL_ref(L, LUA_REGISTRYINDEX);
+    ctx->cron_next = INT_MAX;
 
     // set up the shared state VM. Used by short-lock events (counters/state)
     // for global visibility.
@@ -236,6 +406,7 @@ void *proxy_init(bool use_uring, bool proxy_memprofile) {
     // constantly fetch them from registry.
     lua_newtable(ctx->proxy_sharedvm); // fgen count
     lua_newtable(ctx->proxy_sharedvm); // fgen slot count
+    lua_newtable(ctx->proxy_sharedvm); // backend down status
 
     // Create/start the IO thread, which we need before servers
     // start getting created.
@@ -286,7 +457,8 @@ void proxy_thread_init(void *ctx, LIBEVENT_THREAD *thr) {
     // We can't use GCGEN until we manage pools with reference counting, as
     // they may never hit GC and thus never release their connection
     // resources.
-    lua_gc(L, LUA_GCINC, 199, 100, 13);
+    lua_gc(L, LUA_GCINC, 199, 100, 12);
+    lua_gc(L, LUA_GCSTOP); // handle GC on our own schedule.
     thr->L = L;
     luaL_openlibs(L);
     proxy_register_libs(ctx, thr, L);
@@ -295,9 +467,9 @@ void proxy_thread_init(void *ctx, LIBEVENT_THREAD *thr) {
         thr->proxy_rng[x] = rand();
     }
 
-    thr->proxy_gc_timer = evtimer_new(thr->base, proxy_gc_poke, thr);
+    thr->proxy_gc_timer = evtimer_new(thr->base, proxy_gc_timerpoke, thr);
     // kick off the timer loop.
-    proxy_gc_poke(0, 0, thr);
+    proxy_gc_timerpoke(0, 0, thr);
 
     // Create a proxy event thread structure to piggyback on the worker.
     proxy_event_thread_t *t = calloc(1, sizeof(proxy_event_thread_t));
@@ -331,15 +503,15 @@ void proxy_submit_cb(io_queue_t *q) {
         mcp_backend_t *be;
         P_DEBUG("%s: queueing req for backend: %p\n", __func__, (void *)p);
         if (p->qcount_incr) {
-            // funny workaround: awaiting IOP's don't count toward
-            // resuming a connection, only the completion of the await
+            // funny workaround: async IOP's don't count toward
+            // resuming a connection, only the completion of the async
             // condition.
             q->count++;
         }
 
-        if (p->await_background) {
-            P_DEBUG("%s: fast-returning await_background object: %p\n", __func__, (void *)p);
-            // intercept await backgrounds
+        if (p->background) {
+            P_DEBUG("%s: fast-returning background object: %p\n", __func__, (void *)p);
+            // intercept background requests
             // this call cannot recurse if we're on the worker thread,
             // since the worker thread has to finish executing this
             // function in order to pick up the returned IO.
@@ -354,6 +526,7 @@ void proxy_submit_cb(io_queue_t *q) {
         } else {
             // emulate some of handler_dequeue()
             STAILQ_INSERT_HEAD(&be->io_head, p, io_next);
+            assert(be->depth > -1);
             be->depth++;
             if (!be->stacked) {
                 be->stacked = true;
@@ -404,26 +577,15 @@ void proxy_submit_cb(io_queue_t *q) {
     return;
 }
 
-// This function handles return processing for the "old style" API: direct
-// pool calls and mcp.await()
+// This function handles return processing for the "old style" API:
+// currently just `mcp.internal()`
 void proxy_return_rctx_cb(io_pending_t *pending) {
     io_pending_proxy_t *p = (io_pending_proxy_t *)pending;
     if (p->client_resp && p->client_resp->blen) {
         // FIXME: workaround for buffer memory being external to objects.
         // can't run 0 since that means something special (run the GC)
         unsigned int kb = p->client_resp->blen / 1000;
-        lua_gc(p->rctx->Lc, LUA_GCSTEP, kb > 0 ? kb : 1);
-    }
-
-    if (p->is_await) {
-        p->rctx->async_pending--;
-        mcplib_await_return(p);
-        // need to directly attempt to return the context,
-        // we may or may not be hitting proxy_run_rcontext from await_return.
-        if (p->rctx->async_pending == 0) {
-            mcp_funcgen_return_rctx(p->rctx);
-        }
-        return;
+        p->thread->proxy_vm_extra_kb += kb > 0 ? kb : 1;
     }
 
     mcp_rcontext_t *rctx = p->rctx;
@@ -529,14 +691,12 @@ int try_read_command_proxy(conn *c) {
 // Must only be called with an active coroutine.
 void proxy_cleanup_conn(conn *c) {
     assert(c->proxy_rctx);
-    LIBEVENT_THREAD *thr = c->thread;
     mcp_rcontext_t *rctx = c->proxy_rctx;
     assert(rctx->pending_reqs == 1);
     rctx->pending_reqs = 0;
 
     mcp_funcgen_return_rctx(rctx);
     c->proxy_rctx = NULL;
-    WSTAT_DECR(thr, proxy_req_active, 1);
 }
 
 // we buffered a SET of some kind.
@@ -669,8 +829,10 @@ static void _proxy_run_tresp_to_resp(mc_resp *tresp, mc_resp *resp) {
     // So far all we fill is the wbuf and some iov's? so just copy
     // that + the UDP info?
     memcpy(resp->wbuf, tresp->wbuf, tresp->iov[0].iov_len);
+    resp->tosend = 0;
     for (int x = 0; x < tresp->iovcnt; x++) {
         resp->iov[x] = tresp->iov[x];
+        resp->tosend += tresp->iov[x].iov_len;
     }
     // resp->iov[x].iov_base needs to be updated if it's
     // pointing within its wbuf.
@@ -689,6 +851,7 @@ static void _proxy_run_tresp_to_resp(mc_resp *tresp, mc_resp *resp) {
     resp->request_addr = tresp->request_addr;
     resp->request_addr_size = tresp->request_addr_size;
     resp->item = tresp->item; // will be populated if not extstore fetch
+    tresp->item = NULL; // move ownership of the item to resp from tresp
     resp->skip = tresp->skip;
 }
 
@@ -703,10 +866,6 @@ static void _proxy_run_tresp_to_resp(mc_resp *tresp, mc_resp *resp) {
 // - need to only increment q->count once per stack of requests coming from a
 //   resp.
 //
-// There are workarounds for this all over. In the await code, we test for
-// "the first await object" or "is an await background object", for
-// incrementing the q->count
-// For pool-backed requests we always increment in submit
 // For RQU backed requests (new API) there isn't an easy place to test for
 // "the first request", because:
 // - The connection queue is a stack of _all_ requests pending on this
@@ -734,15 +893,15 @@ static void _proxy_run_tresp_to_resp(mc_resp *tresp, mc_resp *resp) {
 int proxy_run_rcontext(mcp_rcontext_t *rctx) {
     int nresults = 0;
     lua_State *Lc = rctx->Lc;
-    int cores = lua_resume(Lc, NULL, 1, &nresults);
+    assert(rctx->lua_narg != 0);
+    int cores = lua_resume(Lc, NULL, rctx->lua_narg, &nresults);
+    rctx->lua_narg = 1; // reset to default since not-default is uncommon.
     size_t rlen = 0;
-    conn *c = rctx->c;
     mc_resp *resp = rctx->resp;
 
     if (cores == LUA_OK) {
         // don't touch the result object if we were a sub-context.
         if (!rctx->parent) {
-            WSTAT_DECR(c->thread, proxy_req_active, 1);
             int type = lua_type(Lc, 1);
             mcp_resp_t *r = NULL;
             P_DEBUG("%s: coroutine completed. return type: %d\n", __func__, type);
@@ -752,7 +911,6 @@ int proxy_run_rcontext(mcp_rcontext_t *rctx) {
                     proxy_out_errstring(resp, PROXY_SERVER_ERROR, "backend failure");
                 } else if (r->cresp) {
                     mc_resp *tresp = r->cresp;
-                    assert(c != NULL);
 
                     _proxy_run_tresp_to_resp(tresp, resp);
                     // we let the mcp_resp gc handler free up tresp and any
@@ -789,25 +947,7 @@ int proxy_run_rcontext(mcp_rcontext_t *rctx) {
         lua_pop(Lc, 1);
 
         int res = 0;
-        mcp_request_t *rq = NULL;
-        mcp_backend_t *be = NULL;
-        mcp_resp_t *r = NULL;
         switch (yield_type) {
-            case MCP_YIELD_AWAIT:
-                // called with await context on the stack.
-                rctx->first_queue = false; // HACK: ensure awaits are counted.
-                mcplib_await_run_rctx(rctx);
-                break;
-            case MCP_YIELD_POOL:
-                // TODO (v2): c only used for cache alloc?
-                // pool_call checks the argument already.
-                be = lua_touserdata(Lc, -1);
-                rq = lua_touserdata(Lc, -2);
-                // not using a pre-made res object from this yield type.
-                r = mcp_prep_resobj(Lc, rq, be, c->thread);
-                rctx->first_queue = false; // HACK: ensure poolreqs are counted.
-                mcp_queue_rctx_io(rctx, rq, be, r);
-                break;
             case MCP_YIELD_INTERNAL:
                 // stack should be: rq, res
                 if (rctx->parent) {
@@ -836,6 +976,9 @@ int proxy_run_rcontext(mcp_rcontext_t *rctx) {
                 // requests, so we still need to iterate the full set of qslots.
                 _proxy_run_rcontext_queues(rctx);
                 break;
+            case MCP_YIELD_SLEEP:
+                // Pause coroutine and do nothing. Alarm will resume.
+                break;
             default:
                 abort();
         }
@@ -846,7 +989,6 @@ int proxy_run_rcontext(mcp_rcontext_t *rctx) {
         P_DEBUG("%s: Failed to run coroutine: %s\n", __func__, lua_tostring(Lc, -1));
         LOGGER_LOG(NULL, LOG_PROXYEVENTS, LOGGER_PROXY_ERROR, NULL, lua_tostring(Lc, -1));
         if (!rctx->parent) {
-            WSTAT_DECR(c->thread, proxy_req_active, 1);
             proxy_out_errstring(resp, PROXY_SERVER_ERROR, "lua failure");
         }
         rctx->pending_reqs--;
@@ -1012,13 +1154,11 @@ static void proxy_process_command(conn *c, char *command, size_t cmdlen, bool mu
     WSTAT_L(c->thread);
     istats->counters[pr.command]++;
     c->thread->stats.proxy_conn_requests++;
-    c->thread->stats.proxy_req_active++;
     active_reqs = c->thread->stats.proxy_req_active;
     WSTAT_UL(c->thread);
 
-    if (active_reqs > ctx->active_req_limit) {
+    if (active_reqs >= ctx->active_req_limit) {
         proxy_out_errstring(c->resp, PROXY_SERVER_ERROR, "active request limit reached");
-        WSTAT_DECR(c->thread, proxy_req_active, 1);
         if (pr.vlen != 0) {
             c->sbytes = pr.vlen;
             conn_set_state(c, conn_swallow);
@@ -1030,7 +1170,6 @@ static void proxy_process_command(conn *c, char *command, size_t cmdlen, bool mu
     mcp_rcontext_t *rctx = mcp_funcgen_start(L, hook_ref.ctx, &pr);
     if (rctx == NULL) {
         proxy_out_errstring(c->resp, PROXY_SERVER_ERROR, "lua start failure");
-        WSTAT_DECR(c->thread, proxy_req_active, 1);
         if (pr.vlen != 0) {
             c->sbytes = pr.vlen;
             conn_set_state(c, conn_swallow);
@@ -1041,6 +1180,7 @@ static void proxy_process_command(conn *c, char *command, size_t cmdlen, bool mu
     mcp_set_request(&pr, rctx->request, command, cmdlen);
     rctx->request->ascii_multiget = multiget;
     rctx->c = c;
+    rctx->conn_fd = c->sfd;
     rctx->pending_reqs++; // seed counter with the "main" request
     // remember the top level mc_resp, because further requests on the
     // same connection will replace c->resp.
@@ -1069,7 +1209,6 @@ static void proxy_process_command(conn *c, char *command, size_t cmdlen, bool mu
             // normal cleanup
             lua_settop(L, 0);
             proxy_out_errstring(c->resp, PROXY_SERVER_ERROR, "out of memory");
-            WSTAT_DECR(c->thread, proxy_req_active, 1);
             c->sbytes = pr.vlen;
             conn_set_state(c, conn_swallow);
             return;
@@ -1152,6 +1291,14 @@ mcp_resp_t *mcp_prep_resobj(lua_State *L, mcp_request_t *rq, mcp_backend_t *be,
     return r;
 }
 
+void mcp_resp_set_elapsed(mcp_resp_t *r) {
+    struct timeval end;
+    // stamp the elapsed time into the response object.
+    gettimeofday(&end, NULL);
+    r->elapsed = (end.tv_sec - r->start.tv_sec) * 1000000 +
+        (end.tv_usec - r->start.tv_usec);
+}
+
 // Used for any cases where we're queueing requests to the IO subsystem.
 // NOTE: it's not currently possible to limit the memory used by the IO
 // object cache. So this check is redundant, and any callers may proceed
@@ -1177,7 +1324,7 @@ io_pending_proxy_t *mcp_queue_rctx_io(mcp_rcontext_t *rctx, mcp_request_t *rq, m
     p->c = c;
     p->client_resp = r;
     p->flushed = false;
-    p->return_cb = proxy_return_rctx_cb;
+    p->return_cb = NULL;
     p->finalize_cb = proxy_finalize_rctx_cb;
 
     // pass along the request context for resumption.
@@ -1229,6 +1376,70 @@ void mcp_sharedvm_delta(proxy_ctx_t *ctx, int tidx, const char *name, int delta)
     pthread_mutex_unlock(&ctx->sharedvm_lock);
 }
 
+void mcp_sharedvm_remove(proxy_ctx_t *ctx, int tidx, const char *name) {
+    lua_State *L = ctx->proxy_sharedvm;
+    pthread_mutex_lock(&ctx->sharedvm_lock);
+
+    lua_pushnil(L);
+    lua_setfield(L, tidx, name);
+
+    pthread_mutex_unlock(&ctx->sharedvm_lock);
+}
+
+// Global object support code.
+// Global objects are created in the configuration VM, and referenced in
+// worker VMs via proxy objects that refer back to memory in the
+// configuration VM.
+// We manage reference counts: once all remote proxy objects are collected, we
+// signal the config thread to remove a final reference and collect garbage to
+// remove the global object.
+
+static void mcp_gobj_enqueue(proxy_ctx_t *ctx, struct mcp_globalobj_s *g) {
+    pthread_mutex_lock(&ctx->manager_lock);
+    STAILQ_INSERT_TAIL(&ctx->manager_head, g, next);
+    pthread_cond_signal(&ctx->manager_cond);
+    pthread_mutex_unlock(&ctx->manager_lock);
+}
+
+// References the object, initializing the self-reference if necessary.
+// Call from config thread, with global object on top of stack.
+void mcp_gobj_ref(lua_State *L, struct mcp_globalobj_s *g) {
+    pthread_mutex_lock(&g->lock);
+    if (g->self_ref == 0) {
+        // Initialization requires a small dance:
+        // - store a negative of our ref, increase refcount an extra time
+        // - then link and signal the manager thread as though we were GC'ing
+        // the object.
+        // - the manager thread will later acknowledge the initialization of
+        // this global object and negate the self_ref again
+        // - this prevents an unused proxy object from causing the global
+        // object to be reaped early while we are still copying it to worker
+        // threads, as the manager thread will block waiting for the config
+        // thread to finish its reload work.
+        g->self_ref = -luaL_ref(L, LUA_REGISTRYINDEX);
+        g->refcount++;
+        proxy_ctx_t *ctx = PROXY_GET_CTX(L);
+        mcp_gobj_enqueue(ctx, g);
+    } else {
+        lua_pop(L, 1); // drop the reference we didn't end up using.
+    }
+    g->refcount++;
+    pthread_mutex_unlock(&g->lock);
+}
+
+void mcp_gobj_unref(proxy_ctx_t *ctx, struct mcp_globalobj_s *g) {
+    pthread_mutex_lock(&g->lock);
+    g->refcount--;
+    if (g->refcount == 0) {
+        mcp_gobj_enqueue(ctx, g);
+    }
+    pthread_mutex_unlock(&g->lock);
+}
+
+void mcp_gobj_finalize(struct mcp_globalobj_s *g) {
+    pthread_mutex_destroy(&g->lock);
+}
+
 static void *mcp_profile_alloc(void *ud, void *ptr, size_t osize,
                                             size_t nsize) {
     struct mcp_memprofile *prof = ud;
@@ -1315,11 +1526,15 @@ __attribute__((unused)) void dump_stack(lua_State *L, const char *msg) {
     fprintf(stderr, "--TOP OF STACK [%d] | %s\n", top, msg);
     for (; i < top + 1; i++) {
         int type = lua_type(L, i);
+        void *udata = NULL;
         // lets find the metatable of this userdata to identify it.
         if (lua_getmetatable(L, i) != 0) {
             lua_pushstring(L, "__name");
             if (lua_rawget(L, -2) != LUA_TNIL) {
-                fprintf(stderr, "--|%d| [%s] (%s)\n", i, lua_typename(L, type), lua_tostring(L, -1));
+                if (type == LUA_TUSERDATA) {
+                    udata = lua_touserdata(L, i);
+                }
+                fprintf(stderr, "--|%d| [%s] (%s) [ptr: %p]\n", i, lua_typename(L, type), lua_tostring(L, -1), udata);
                 lua_pop(L, 2);
                 continue;
             }
@@ -1328,7 +1543,10 @@ __attribute__((unused)) void dump_stack(lua_State *L, const char *msg) {
         if (type == LUA_TSTRING) {
             fprintf(stderr, "--|%d| [%s] | %s\n", i, lua_typename(L, type), lua_tostring(L, i));
         } else {
-            fprintf(stderr, "--|%d| [%s]\n", i, lua_typename(L, type));
+            if (type == LUA_TUSERDATA) {
+                udata = lua_touserdata(L, i);
+            }
+            fprintf(stderr, "--|%d| [%s] [ptr: %p]\n", i, lua_typename(L, type), udata);
         }
     }
     fprintf(stderr, "-----------------\n");
@@ -1373,3 +1591,85 @@ __attribute__((unused)) void dump_registry(lua_State *L, const char *msg) {
     fprintf(stderr, "### TABLE\t[%d]\n", table);
     fprintf(stderr, "-----------------\n");
 }
+
+// Searches for a function generator with a specific name attached.
+// Adding breakpoints on the print lines lets you inspect the fgen and its
+// slots.
+__attribute__((unused)) void dump_funcgen(lua_State *L, const char *name, const char *msg) {
+    int ref_size = lua_rawlen(L, LUA_REGISTRYINDEX);
+    fprintf(stderr, "--LUA FUNCGEN FINDER [%d] | %s\n", ref_size, msg);
+    // walk registry
+    int ridx = lua_absindex(L, LUA_REGISTRYINDEX);
+    lua_pushnil(L);
+    while (lua_next(L, ridx) != 0) {
+        int type = lua_type(L, -1);
+        if (type == LUA_TUSERDATA) {
+            mcp_funcgen_t *f = luaL_testudata(L, -1, "mcp.funcgen");
+            if (f != NULL && strcmp(name, f->name) == 0) {
+                fprintf(stderr, "===found funcgen [%s] [%p]===\n", f->name, (void *)f);
+                lua_getiuservalue(L, -1, 1);
+                int tidx = lua_absindex(L, -1);
+                lua_pushnil(L);
+                while (lua_next(L, tidx) != 0) {
+                    mcp_rcontext_t *rctx = lua_touserdata(L, -1);
+                    if (rctx != NULL) {
+                        fprintf(stderr, "-- slot: [%p]\n", (void *)rctx);
+                    }
+                    lua_pop(L, 1); // drop value
+                }
+                lua_pop(L, 1); // drop slot table
+            }
+        }
+        lua_pop(L, 1); // drop value
+    }
+    fprintf(stderr, "-----------------\n");
+}
+
+static void dump_pool_info(mcp_pool_t *p) {
+    fprintf(stderr, "--pool: [%s] size: [%d] be_total: [%d] rc: [%d] io: [%d]\n",
+            p->beprefix, p->pool_size, p->pool_be_total, p->g.refcount, p->use_iothread);
+
+    for (int x = 0; x < p->pool_be_total; x++) {
+        mcp_backend_t *be = p->pool[x].be;
+        // Dumb: pool_be_total is wrong if pool is using iothread. Why?
+        if (be != NULL) {
+            fprintf(stderr, "  --be[%d] label: [%s] name: [%s] conns: [%d] depth: [%d]\n",
+                    x, be->label, be->name, be->conncount, be->depth);
+            for (int i = 0; i < be->conncount; i++) {
+                struct mcp_backendconn_s *bec = &be->be[i];
+                fprintf(stderr, "    --bec[%d] bad: [%d] failcnt: [%d] depth: [%d] state: [%d] can_write[%d] write_event[%d]\n",
+                        i, bec->bad, bec->failed_count, bec->depth, bec->state, bec->can_write, event_pending(&bec->timeout_event, EV_WRITE, NULL));
+            }
+        }
+    }
+    fprintf(stderr, "=======\n");
+}
+
+// Dumps some info about pools.
+// If given the config thread, it should find the main pools
+// If given a worker thread, it will look for the pool proxy objects and find
+// the main pools that way.
+__attribute__((unused)) void dump_pools(lua_State *L, const char *msg) {
+    int ref_size = lua_rawlen(L, LUA_REGISTRYINDEX);
+    fprintf(stderr, "--LUA POOL DUMPER [%d] | %s\n", ref_size, msg);
+    // walk registry
+    int ridx = lua_absindex(L, LUA_REGISTRYINDEX);
+    lua_pushnil(L);
+    while (lua_next(L, ridx) != 0) {
+        int type = lua_type(L, -1);
+        if (type == LUA_TUSERDATA) {
+            mcp_pool_t *p = luaL_testudata(L, -1, "mcp.pool");
+            if (p != NULL) {
+                dump_pool_info(p);
+            } else {
+                mcp_pool_proxy_t *pp = luaL_testudata(L, -1, "mcp.pool_proxy");
+                if (pp != NULL) {
+                    dump_pool_info(pp->main);
+                }
+            }
+        }
+        lua_pop(L, 1); // drop value
+    }
+    fprintf(stderr, "-----------------\n");
+
+}
