@@ -88,6 +88,10 @@ struct mcp_memprofile {
     uint64_t alloc_bytes[8];
 };
 
+// for various time conversion functions
+#define NANOSECONDS(x) ((x) * 1E9 + 0.5)
+#define MICROSECONDS(x) ((x) * 1E6 + 0.5)
+
 // Note: value created from thin air. Could be shorter.
 #define MCP_REQUEST_MAXLEN KEY_MAX_LENGTH * 2
 
@@ -96,14 +100,14 @@ struct mcp_memprofile {
 
 #define MCP_BACKEND_UPVALUE 1
 
-#define MCP_YIELD_POOL 1
-#define MCP_YIELD_AWAIT 2
-#define MCP_YIELD_INTERNAL 3
-#define MCP_YIELD_WAITCOND 4
-#define MCP_YIELD_WAITHANDLE 5
+#define MCP_YIELD_INTERNAL 1
+#define MCP_YIELD_WAITCOND 2
+#define MCP_YIELD_WAITHANDLE 3
+#define MCP_YIELD_SLEEP 4
 
 #define SHAREDVM_FGEN_IDX 1
 #define SHAREDVM_FGENSLOT_IDX 2
+#define SHAREDVM_BACKEND_IDX 3
 
 // all possible commands.
 #define CMD_FIELDS \
@@ -179,25 +183,31 @@ void *proxy_event_thread_ur(void *arg);
 // Note: This ends up wasting a few counters, but simplifies the rest of the
 // process for handling internal worker stats.
 struct proxy_int_stats {
+    uint64_t vm_gc_runs;
+    uint64_t vm_memory_kb;
     uint64_t counters[CMD_FINAL];
 };
 
 struct proxy_user_stats {
-    size_t num_stats; // number of stats, for sizing various arrays
-    char **names; // not needed for worker threads
+    int num_stats; // number of stats, for sizing various arrays
     uint64_t *counters; // array of counters.
 };
 
+struct proxy_user_stats_entry {
+    char *name;
+    unsigned int cname; // offset into compact name buffer
+    bool reset; // counter must reset this cycle
+};
+
 struct proxy_global_stats {
     uint64_t config_reloads;
     uint64_t config_reload_fails;
+    uint64_t config_cron_runs;
+    uint64_t config_cron_fails;
     uint64_t backend_total;
-    uint64_t backend_disconn; // backends with no connections
-    uint64_t backend_requests; // reqs sent to backends
-    uint64_t backend_responses; // responses received from backends
-    uint64_t backend_errors; // errors from backends
     uint64_t backend_marked_bad; // backend set to autofail
     uint64_t backend_failed; // an error caused a backend reset
+    uint64_t request_failed_depth; // requests fast-failed due to be depth
 };
 
 struct proxy_tunables {
@@ -207,21 +217,33 @@ struct proxy_tunables {
     struct timeval flap; // need to stay connected this long or it's flapping
     float flap_backoff_ramp; // factorial for retry time
     uint32_t flap_backoff_max; // don't backoff longer than this.
+    int backend_depth_limit; // requests fast fail once depth over this limit
     int backend_failure_limit;
     int max_ustats; // limit the ustats index.
     bool tcp_keepalive;
     bool use_iothread; // default for using the bg io thread.
+    bool use_tls; // whether or not be should use TLS
     bool down; // backend is forced into a down/bad state.
 };
 
 typedef STAILQ_HEAD(globalobj_head_s, mcp_globalobj_s) globalobj_head_t;
 typedef struct {
     lua_State *proxy_state; // main configuration vm
-    lua_State *proxy_sharedvm; // sub VM for short-lock global events/data
-    void *proxy_code;
     proxy_event_thread_t *proxy_io_thread;
     uint64_t active_req_limit; // max total in-flight requests
     uint64_t buffer_memory_limit; // max bytes for send/receive buffers.
+#ifdef PROXY_TLS
+    void *tls_ctx;
+#endif
+    int user_stats_num; // highest seen stat index
+    struct proxy_user_stats_entry *user_stats;
+    char *user_stats_namebuf; // compact linear buffer for stat names
+    struct proxy_tunables tunables; // NOTE: updates covered by stats_lock
+    struct proxy_global_stats global_stats;
+    // less frequently used entries down here.
+    void *proxy_code;
+    lua_State *proxy_sharedvm; // sub VM for short-lock global events/data
+    pthread_mutex_t stats_lock; // used for rare global counters
     pthread_mutex_t config_lock;
     pthread_cond_t config_cond;
     pthread_t config_tid;
@@ -232,16 +254,15 @@ typedef struct {
     pthread_cond_t manager_cond;
     pthread_mutex_t sharedvm_lock; // protect statevm above
     globalobj_head_t manager_head; // stack for pool deallocation.
+    int config_generation; // counter tracking config reloads
+    int cron_ref; // reference to lua cron table
+    int cron_next; // next cron to sleep to / execute
     bool worker_done; // signal variable for the worker lock/cond system.
     bool worker_failed; // covered by worker_lock as well.
     bool use_uring; // use IO_URING for backend connections.
     bool loading; // bool indicating an active config load.
     bool memprofile; // indicate if we want to profile lua memory.
     uint8_t memprofile_thread_counter;
-    struct proxy_global_stats global_stats;
-    struct proxy_user_stats user_stats;
-    struct proxy_tunables tunables; // NOTE: updates covered by stats_lock
-    pthread_mutex_t stats_lock; // used for rare global counters
 } proxy_ctx_t;
 
 #define PROXY_GET_THR_CTX(L) ((*(LIBEVENT_THREAD **)lua_getextraspace(L))->proxy_ctx)
@@ -287,6 +308,7 @@ enum mcp_backend_states {
     mcp_backend_next_close, // complete current request, then close socket
 };
 
+typedef struct mcp_cron_s mcp_cron_t;
 typedef struct mcp_backend_wrap_s mcp_backend_wrap_t;
 typedef struct mcp_backend_label_s mcp_backend_label_t;
 typedef struct mcp_backend_s mcp_backend_t;
@@ -332,6 +354,13 @@ struct mcp_request_s {
     char request[];
 };
 
+struct mcp_cron_s {
+    uint32_t gen;
+    uint32_t next;
+    uint32_t every;
+    bool repeat;
+};
+
 typedef STAILQ_HEAD(io_head_s, _io_pending_proxy_t) io_head_t;
 #define MAX_LABELLEN 512
 #define MAX_NAMELEN 255
@@ -339,7 +368,9 @@ typedef STAILQ_HEAD(io_head_s, _io_pending_proxy_t) io_head_t;
 // TODO (v2): IOV_MAX tends to be 1000+ which would allow for more batching but we
 // don't have a good temporary space and don't want to malloc/free on every
 // write. transmit() uses the stack but we can't do that for uring's use case.
-#if (IOV_MAX > 1024)
+#if MEMCACHED_DEBUG
+#define BE_IOV_MAX 128 // let bench tests trigger max condition easily
+#elif (IOV_MAX > 1024)
 #define BE_IOV_MAX 1024
 #else
 #define BE_IOV_MAX IOV_MAX
@@ -372,6 +403,9 @@ struct mcp_backendconn_s {
     int flap_count; // number of times we've "flapped" into bad state.
     proxy_event_thread_t *event_thread; // event thread owning this backend.
     void *client; // mcmc client
+#ifdef PROXY_TLS
+    void *ssl;
+#endif
     io_head_t io_head; // stack of requests.
     io_pending_proxy_t *io_next; // next request to write.
     char *rbuf; // statically allocated read buffer.
@@ -387,6 +421,9 @@ struct mcp_backendconn_s {
     bool validating; // in process of validating a new backend connection.
     bool can_write; // recently got a WANT_WRITE or are connecting.
     bool bad; // timed out, marked as bad.
+#ifndef PROXY_TLS
+    bool ssl;
+#endif
     struct iovec write_iovs[BE_IOV_MAX]; // iovs to stage batched writes
 };
 
@@ -422,6 +459,10 @@ struct proxy_event_thread_s {
     eventfd_t event_counter;
     eventfd_t beevent_counter;
     bool use_uring;
+#endif
+#ifdef PROXY_TLS
+    char *tls_wbuf;
+    size_t tls_wbuf_size;
 #endif
     pthread_mutex_t mutex; // covers stack.
     pthread_cond_t cond; // condition to wait on while stack drains.
@@ -449,6 +490,7 @@ enum mcp_resp_mode {
 #define RESP_CMD_MAX 8
 typedef struct {
     mcmc_resp_t resp;
+    mcmc_tokenizer_t tok; // optional tokenization of res
     char *buf; // response line + potentially value.
     mc_resp *cresp; // client mc_resp object during extstore fetches.
     LIBEVENT_THREAD *thread; // cresp's owner thread needed for extstore cleanup.
@@ -503,13 +545,9 @@ struct _io_pending_proxy_t {
             struct iovec iov[2]; // request string + tail buffer
             int iovcnt; // 1 or 2...
             unsigned int iovbytes; // total bytes in the iovec
-            int mcpres_ref; // mcp.res reference used for await()
-            int await_ref; // lua reference if we were an await object
             mcp_resp_t *client_resp; // reference (currently pointing to a lua object)
             bool flushed; // whether we've fully written this request to a backend.
-            bool is_await; // are we an await object?
-            bool await_first; // are we the main route for an await object?
-            bool await_background; // dummy IO for backgrounded awaits
+            bool background; // dummy IO for backgrounded awaits
             bool qcount_incr; // HACK.
         };
     };
@@ -556,6 +594,11 @@ typedef struct {
 // utils
 bool proxy_bufmem_checkadd(LIBEVENT_THREAD *t, int len);
 void mcp_sharedvm_delta(proxy_ctx_t *ctx, int tidx, const char *name, int delta);
+void mcp_sharedvm_remove(proxy_ctx_t *ctx, int tidx, const char *name);
+
+void mcp_gobj_ref(lua_State *L, struct mcp_globalobj_s *g);
+void mcp_gobj_unref(proxy_ctx_t *ctx, struct mcp_globalobj_s *g);
+void mcp_gobj_finalize(struct mcp_globalobj_s *g);
 
 // networking interface
 void proxy_init_event_thread(proxy_event_thread_t *t, proxy_ctx_t *ctx, struct event_base *base);
@@ -564,22 +607,9 @@ void proxy_run_backend_queue(be_head_t *head);
 struct mcp_backendconn_s *proxy_choose_beconn(mcp_backend_t *be);
 mcp_resp_t *mcp_prep_resobj(lua_State *L, mcp_request_t *rq, mcp_backend_t *be, LIBEVENT_THREAD *t);
 mcp_resp_t *mcp_prep_bare_resobj(lua_State *L, LIBEVENT_THREAD *t);
+void mcp_resp_set_elapsed(mcp_resp_t *r);
 io_pending_proxy_t *mcp_queue_rctx_io(mcp_rcontext_t *rctx, mcp_request_t *rq, mcp_backend_t *be, mcp_resp_t *r);
 
-// await interface
-enum mcp_await_e {
-    AWAIT_GOOD = 0, // looks for OK + NOT MISS
-    AWAIT_ANY, // any response, including errors,
-    AWAIT_OK, // any non-error response
-    AWAIT_FIRST, // return the result from the first pool
-    AWAIT_FASTGOOD, // returns on first hit or majority non-error
-    AWAIT_BACKGROUND, // returns as soon as background jobs are dispatched
-};
-int mcplib_await(lua_State *L);
-int mcplib_await_logerrors(lua_State *L);
-int mcplib_await_run_rctx(mcp_rcontext_t *rctx);
-int mcplib_await_return(io_pending_proxy_t *p);
-
 // internal request interface
 int mcplib_internal(lua_State *L);
 int mcplib_internal_run(mcp_rcontext_t *rctx);
@@ -612,11 +642,34 @@ enum mcp_rqueue_e {
     QWAIT_OK,
     QWAIT_GOOD,
     QWAIT_FASTGOOD,
-    QWAIT_HANDLE
+    QWAIT_HANDLE,
+    QWAIT_SLEEP,
+};
+
+#define FGEN_NAME_MAXLEN 80
+struct mcp_funcgen_s {
+    LIBEVENT_THREAD *thread; // worker thread that created this funcgen.
+    int generator_ref; // reference to the generator function.
+    int self_ref; // self-reference if we're attached anywhere
+    int argument_ref; // reference to an argument to pass to generator
+    int max_queues; // how many queue slots rctx's have
+    unsigned int refcount; // reference counter
+    unsigned int total; // total contexts managed
+    unsigned int free; // free contexts
+    unsigned int free_max; // size of list below.
+    unsigned int free_pressure; // "pressure" for when to early release rctx
+    bool closed; // the hook holding this fgen has been replaced
+    bool ready; // if we're locked down or not.
+    bool is_router; // if this fgen is actually a router object.
+    struct timespec free_waiter; // must be "too free" for this much time
+    mcp_rcontext_t **list;
+    struct mcp_rqueue_s *queue_list;
+    char name[FGEN_NAME_MAXLEN+1]; // string name for the generator.
 };
 
 enum mcp_funcgen_router_e {
     FGEN_ROUTER_NONE = 0,
+    FGEN_ROUTER_CMDMAP,
     FGEN_ROUTER_SHORTSEP,
     FGEN_ROUTER_LONGSEP,
     FGEN_ROUTER_ANCHORSM,
@@ -628,7 +681,12 @@ struct mcp_router_long_s {
     char stop[KEY_HASH_FILTER_MAX+1];
 };
 
+// To simplify the attach/start code we wrap a funcgen with the router
+// structure. This allows us to have a larger router structure without
+// bloating the fgen object itself, and still benefit from letting funcgen
+// new/cleanup handle most of the memory management.
 struct mcp_funcgen_router {
+    mcp_funcgen_t fgen_self;
     enum mcp_funcgen_router_e type;
     union {
         char sep;
@@ -636,27 +694,9 @@ struct mcp_funcgen_router {
         char anchorsm[2]; // short anchored mode.
         struct mcp_router_long_s big;
     } conf;
-    mcp_funcgen_t *def_fgen; // default route
     int map_ref;
-};
-
-struct mcp_funcgen_s {
-    LIBEVENT_THREAD *thread; // worker thread that created this funcgen.
-    int generator_ref; // reference to the generator function.
-    int self_ref; // self-reference if we're attached anywhere
-    int argument_ref; // reference to an argument to pass to generator
-    int name_ref; // reference to string name for the generator
-    int max_queues; // how many queue slots rctx's have
-    unsigned int refcount; // reference counter
-    unsigned int total; // total contexts managed
-    unsigned int free; // free contexts
-    unsigned int free_max; // size of list below.
-    unsigned int routecount; // total routes if this fgen is a router.
-    bool closed; // the hook holding this fgen has been replaced
-    bool ready; // if we're locked down or not.
-    mcp_rcontext_t **list;
-    struct mcp_rqueue_s *queue_list;
-    struct mcp_funcgen_router router;
+    mcp_funcgen_t *def_fgen; // default route
+    mcp_funcgen_t *cmap[CMD_END_STORAGE]; // fallback command map
 };
 
 #define RQUEUE_TYPE_NONE 0
@@ -691,6 +731,7 @@ struct mcp_rqueue_s {
 };
 
 struct mcp_rcontext_s {
+    int self_ref; // reference to our own object
     int request_ref; // top level request for this context.
     int function_ref; // ref to the created route function.
     int coroutine_ref; // ref to our encompassing coroutine.
@@ -700,7 +741,9 @@ struct mcp_rcontext_s {
     unsigned int wait_done; // TODO: change these variables to uint8's
     int wait_handle; // waiting on a specific queue slot
     int parent_handle; // queue slot in parent rctx
+    int conn_fd; // fd of the originating client, as *c can become invalid
     enum mcp_rqueue_e wait_mode;
+    uint8_t lua_narg; // number of responses to push when yield resuming.
     bool first_queue; // HACK
     lua_State *Lc; // coroutine thread pointer.
     mcp_request_t *request; // ptr to the above reference.
@@ -708,9 +751,12 @@ struct mcp_rcontext_s {
     conn *c; // associated client object.
     mc_resp *resp; // top level response object to fill.
     mcp_funcgen_t *fgen; // parent function generator context.
+    struct event timeout_event; // for *_wait_timeout() and sleep() calls
     struct mcp_rqueue_s qslots[]; // queueable slots.
 };
 
+#define mcp_is_flag_invalid(f) (f < 65 || f > 122)
+
 void mcp_run_rcontext_handle(mcp_rcontext_t *rctx, int handle);
 void mcp_process_rctx_wait(mcp_rcontext_t *rctx, int handle);
 int mcp_process_rqueue_return(mcp_rcontext_t *rctx, int handle, mcp_resp_t *res);
@@ -723,6 +769,11 @@ int mcplib_rcontext_res_good(lua_State *L);
 int mcplib_rcontext_res_any(lua_State *L);
 int mcplib_rcontext_res_ok(lua_State *L);
 int mcplib_rcontext_result(lua_State *L);
+int mcplib_rcontext_cfd(lua_State *L);
+int mcplib_rcontext_tls_peer_cn(lua_State *L);
+int mcplib_rcontext_request_new(lua_State *L);
+int mcplib_rcontext_response_new(lua_State *L);
+int mcplib_rcontext_sleep(lua_State *L);
 int mcplib_funcgenbare_new(lua_State *L);
 int mcplib_funcgen_new(lua_State *L);
 int mcplib_funcgen_new_handle(lua_State *L);
@@ -746,15 +797,44 @@ int mcplib_request_key(lua_State *L);
 int mcplib_request_ltrimkey(lua_State *L);
 int mcplib_request_rtrimkey(lua_State *L);
 int mcplib_request_token(lua_State *L);
+int mcplib_request_token_int(lua_State *L);
 int mcplib_request_ntokens(lua_State *L);
 int mcplib_request_has_flag(lua_State *L);
 int mcplib_request_flag_token(lua_State *L);
+int mcplib_request_flag_token_int(lua_State *L);
+int mcplib_request_flag_add(lua_State *L);
+int mcplib_request_flag_set(lua_State *L);
+int mcplib_request_flag_replace(lua_State *L);
+int mcplib_request_flag_del(lua_State *L);
 int mcplib_request_gc(lua_State *L);
+int mcplib_request_match_res(lua_State *L);
 void mcp_request_cleanup(LIBEVENT_THREAD *t, mcp_request_t *rq);
 
 // response interface
+int mcplib_response_elapsed(lua_State *L);
+int mcplib_response_ok(lua_State *L);
+int mcplib_response_hit(lua_State *L);
+int mcplib_response_vlen(lua_State *L);
+int mcplib_response_code(lua_State *L);
+int mcplib_response_line(lua_State *L);
+int mcplib_response_flag_blank(lua_State *L);
+
+// inspector interface
+int mcplib_req_inspector_new(lua_State *L);
+int mcplib_res_inspector_new(lua_State *L);
+int mcplib_inspector_gc(lua_State *L);
+int mcplib_inspector_call(lua_State *L);
+
+// mutator interface
+int mcplib_req_mutator_new(lua_State *L);
+int mcplib_res_mutator_new(lua_State *L);
+int mcplib_mutator_gc(lua_State *L);
+int mcplib_mutator_call(lua_State *L);
+
 void mcp_response_cleanup(LIBEVENT_THREAD *t, mcp_resp_t *r);
 void mcp_set_resobj(mcp_resp_t *r, mcp_request_t *rq, mcp_backend_t *be, LIBEVENT_THREAD *t);
+int mcplib_response_gc(lua_State *L);
+int mcplib_response_close(lua_State *L);
 
 int mcplib_open_dist_jump_hash(lua_State *L);
 int mcplib_open_dist_ring_hash(lua_State *L);
@@ -762,7 +842,11 @@ int mcplib_open_dist_ring_hash(lua_State *L);
 int proxy_run_rcontext(mcp_rcontext_t *rctx);
 mcp_backend_t *mcplib_pool_proxy_call_helper(mcp_pool_proxy_t *pp, const char *key, size_t len);
 void mcp_request_attach(mcp_request_t *rq, io_pending_proxy_t *p);
-int mcp_request_render(mcp_request_t *rq, int idx, const char *tok, size_t len);
+int mcp_request_render(mcp_request_t *rq, int idx, char flag, const char *tok, size_t len);
+int mcp_request_append(mcp_request_t *rq, const char flag, const char *tok, size_t len);
+int mcp_request_find_flag_index(mcp_request_t *rq, const char flag);
+int mcp_request_find_flag_token(mcp_request_t *rq, const char flag, const char **token, size_t *len);
+int mcp_request_find_flag_tokenint64(mcp_request_t *rq, const char flag, int64_t *token);
 void proxy_lua_error(lua_State *L, const char *s);
 #define proxy_lua_ferror(L, fmt, ...) \
     do { \
@@ -781,4 +865,6 @@ int mcplib_open_hash_xxhash(lua_State *L);
 
 __attribute__((unused)) void dump_stack(lua_State *L, const char *msg);
 __attribute__((unused)) void dump_registry(lua_State *L, const char *msg);
+__attribute__((unused)) void dump_funcgen(lua_State *L, const char *name, const char *msg);
+__attribute__((unused)) void dump_pools(lua_State *L, const char *msg);
 #endif
