@@ -487,7 +487,7 @@ static void process_delete_cmd(LIBEVENT_THREAD *t, mcp_parser_t *pr, mc_resp *re
         pthread_mutex_lock(&t->stats.mutex);
         t->stats.slab_stats[ITEM_clsid(it)].delete_hits++;
         pthread_mutex_unlock(&t->stats.mutex);
-
+        LOGGER_LOG(NULL, LOG_DELETIONS, LOGGER_DELETIONS, it, LOG_TYPE_DELETE);
         do_item_unlink(it, hv);
         STORAGE_delete(t->storage, it);
         do_item_remove(it);      /* release our reference */
@@ -1367,6 +1367,8 @@ static void process_mdelete_cmd(LIBEVENT_THREAD *t, mcp_parser_t *pr, mc_resp *r
             pthread_mutex_lock(&t->stats.mutex);
             t->stats.slab_stats[ITEM_clsid(it)].delete_hits++;
             pthread_mutex_unlock(&t->stats.mutex);
+            LOGGER_LOG(NULL, LOG_DELETIONS, LOGGER_DELETIONS, it, LOG_TYPE_META_DELETE);
+
             if (!of.remove_val) {
                 do_item_unlink(it, hv);
                 STORAGE_delete(t->storage, it);
@@ -1579,6 +1581,7 @@ static void process_marithmetic_cmd(LIBEVENT_THREAD *t, mcp_parser_t *pr, mc_res
                     }
                     META_SPACE(p);
                     memcpy(p, &pr->request[pr->tokens[i]], tlen);
+                    p += tlen;
                     break;
                 case 'k':
                     META_KEY(p, key, nkey, of.key_binary);
@@ -1607,6 +1610,7 @@ static void process_marithmetic_cmd(LIBEVENT_THREAD *t, mcp_parser_t *pr, mc_res
                     }
                     META_SPACE(p);
                     memcpy(p, &pr->request[pr->tokens[i]], tlen);
+                    p += tlen;
                     break;
                 case 'k':
                     META_KEY(p, key, nkey, of.key_binary);
@@ -1724,7 +1728,7 @@ int mcplib_internal_run(mcp_rcontext_t *rctx) {
     // wrap the main commands with something that decorates r->resp directly
     // instead of going through a parser to save some CPU.
     // Either way this is a lot less code.
-    mcmc_bare_parse_buf(resp->iov[0].iov_base, resp->iov[0].iov_len, &r->resp);
+    mcmc_parse_buf(resp->iov[0].iov_base, resp->iov[0].iov_len, &r->resp);
 
     if (rq->ascii_multiget) {
         if (r->resp.type == MCMC_RESP_GET) {
